@@ -25,8 +25,10 @@
 
 #include <stdio.h>
 #include <fcntl.h>
-#include <sys/stat.h>
+#include <dirent.h>
+#include <string.h>
 #include <gdbm.h>
+#include <sys/stat.h>
 #include "yp.h"
 
 
@@ -58,12 +60,23 @@ int is_valid_domain(char *name)
 bool_t *ypproc_domain_2_svc(domainname *name, struct svc_req *rqstp)
 {
     static bool_t result;
-    
+
+
+    if (debug_flag)
+    {
+	fprintf(stderr, "ypproc_domain(\"%s\")\n",
+		*name);
+    }
+
     if (is_valid_domain(*name))
 	result = TRUE;
     else
 	result = FALSE;
     
+    if (debug_flag)
+	fprintf(stderr, "\t-> %s.\n",
+		(result == TRUE ? "Ok" : "Not served by us"));
+    
     return &result;
 }
 
@@ -71,15 +84,28 @@ bool_t *ypproc_domain_nonack_2_svc(domainname *name, struct svc_req *rqstp)
 {
     static bool_t result;
 
+    if (debug_flag)
+    {
+	fprintf(stderr, "ypproc_domain_nonack(\"%s\")\n",
+		*name);
+    }
+    
     if (!is_valid_domain(*name))
+    {
+	if (debug_flag)
+	    fprintf(stderr, "\t-> Not a valid domain.\n");
 	return NULL;
+    }
+    
+    if (debug_flag)
+	fprintf(stderr, "\t-> OK.\n");
     
     result = TRUE;
     return &result;
 }
 
 
-GDBM_FILE open_database(char *domain, char *map)
+static GDBM_FILE open_database(char *domain, char *map)
 {
     GDBM_FILE dbp;
     char buf[1025];
@@ -94,6 +120,18 @@ GDBM_FILE open_database(char *domain, char *map)
     return dbp;
 }
 
+static unsigned long get_dtm(char *domain, char *map)
+{
+    struct stat sbuf;
+    char buf[1025];
+
+    sprintf(buf, "%s/%s/%s", path_ypdb, domain, map);
+    if (stat(buf, &sbuf) < 0)
+	return 0;
+    else
+	return (unsigned long) sbuf.st_mtime;
+}
+
 
 ypresp_val *ypproc_match_2_svc(ypreq_key *key, struct svc_req *rqstp)
 {
@@ -228,7 +266,7 @@ ypresp_key_val *ypproc_next_2_svc(ypreq_key *key, struct svc_req *rqstp)
 	    
 	    dkey = gdbm_nextkey(dbp, dkey);
 	    if (dkey.dptr == NULL)
-		result.stat = YP_NOKEY;
+		result.stat = YP_NOMORE;
 	    else
 	    {
 		dval = gdbm_fetch(dbp, dkey);
@@ -249,7 +287,7 @@ ypresp_key_val *ypproc_next_2_svc(ypreq_key *key, struct svc_req *rqstp)
     return &result;
 }
 
-print_ypmap_parms(struct ypmap_parms *pp)
+static void print_ypmap_parms(struct ypmap_parms *pp)
 {
     fprintf(stderr, "\t\tdomain   = \"%s\"\n", pp->domain);
     fprintf(stderr, "\t\tmap      = \"%s\"\n", pp->map);
@@ -280,13 +318,26 @@ void *ypproc_clear_2_svc(void *dummy, struct svc_req *rqstp)
 {
     static int foo;
 
+    if (debug_flag)
+    	fprintf(stderr, "ypproc_clear_2_svc()\n");
+
     return (void *) &foo;
 }
 
+
+
+/*
+** Hmm... I wonder how on Earth this one is supposed to work. Perhaps
+** like this?
+*/
 ypresp_all *ypproc_all_2_svc(ypreq_nokey *nokey, struct svc_req *rqstp)
 {
-    static ypresp_all result;
-
+    static ypresp_all *result = NULL;
+    static int reslen = 0;
+    GDBM_FILE dbp;
+    int i;
+    
+    
     if (debug_flag)
     {
 	fprintf(stderr, "ypproc_all_2_svc():\n");
@@ -294,19 +345,114 @@ ypresp_all *ypproc_all_2_svc(ypreq_nokey *nokey, struct svc_req *rqstp)
 	fprintf(stderr, "\tmap = \"%s\"\n", nokey->map);
     }
 
-    result.more = FALSE;
+    if (result == NULL)
+    {
+	result = malloc(sizeof(*result));
+	if (result == NULL)
+	{
+	    /* Fatal error. Should babble to syslog et al... */
+	    return NULL;
+	}
+	reslen = 0;
+    }
 
-    if (!is_valid_domain(nokey->domain))
-	result.ypresp_all_u.val.stat = YP_NODOM;
-    else
-	result.ypresp_all_u.val.stat = YP_BADDB;
-
-    result.ypresp_all_u.val.key.keydat_len = 0;
-    result.ypresp_all_u.val.key.keydat_val = NULL;
-    result.ypresp_all_u.val.val.valdat_len = 0;
-    result.ypresp_all_u.val.val.valdat_val = NULL;
+    i = 0;
+	
     
-    return &result;
+    if ((dbp = open_database(nokey->domain, nokey->map)) == NULL)
+	result->ypresp_all_u.val.stat = YP_NODOM;
+    else
+#if 1
+    {
+	result[i].more = 0;
+	result[i].ypresp_all_u.val.stat = YP_YPERR;
+	result[i].ypresp_all_u.val.key.keydat_len = 0;
+	result[i].ypresp_all_u.val.key.keydat_val = NULL;
+	result[i].ypresp_all_u.val.val.valdat_len = 0;
+	result[i].ypresp_all_u.val.val.valdat_val = NULL;
+    }
+#else
+    {
+	datum key, val;
+
+	for (key = gdbm_firstkey(dbp); key.dptr; key = gdbm_nextkey(dbp, key))
+	{
+	    if (i >= reslen)
+	    {
+		int j;
+
+		reslen += 32;
+		if (debug_flag)
+		    fprintf(stderr, "\t(Reallocating buffer to %d slots)\n",
+			    reslen);
+		
+		result = realloc(result, sizeof(*result)*(reslen+1));
+		if (result == NULL)
+		{
+		    /* Fatal error. Should babble to syslog et al... */
+		    reslen = 0;
+		    gdbm_close(dbp);
+		    return NULL;
+		}
+
+		for (j = i; j <= reslen; j++)
+		{
+		    result[j].more = 0;
+		    result[j].ypresp_all_u.val.key.keydat_len = 0;
+		    result[j].ypresp_all_u.val.key.keydat_val = NULL;
+		    result[j].ypresp_all_u.val.val.valdat_len = 0;
+		    result[j].ypresp_all_u.val.val.valdat_val = NULL;
+		}
+	    }
+	    
+	    if (result[i].ypresp_all_u.val.key.keydat_val)
+	    {
+		free(result[i].ypresp_all_u.val.key.keydat_val);
+		result[i].ypresp_all_u.val.key.keydat_val = NULL;
+	    }
+	    
+	    if (result[i].ypresp_all_u.val.val.valdat_val)
+	    {
+		free(result[i].ypresp_all_u.val.val.valdat_val);
+		result[i].ypresp_all_u.val.val.valdat_val = NULL;
+	    }
+
+	    val = gdbm_fetch(dbp, key);
+
+/* XXXX check val.dptr... */	    
+	    
+	    if (debug_flag)
+		fprintf(stderr,
+			"\t(Inserting at slot %d: key = \"%.*s\", value = \"%.*s\")\n",
+			i, key.dsize, key.dptr, val.dsize, val.dptr);
+			
+	    result[i].more = 1;
+	    result[i].ypresp_all_u.val.stat = YP_TRUE;
+	    result[i].ypresp_all_u.val.key.keydat_len = key.dsize;
+	    result[i].ypresp_all_u.val.key.keydat_val = strdup(key.dptr);
+	    result[i].ypresp_all_u.val.val.valdat_len = val.dsize;
+	    result[i].ypresp_all_u.val.val.valdat_val = strdup(val.dptr);
+
+	    i++;
+	}
+	     
+	gdbm_close(dbp);
+    }
+    
+    if (debug_flag)
+	fprintf(stderr, "\t-> Returning %d values\n", i);
+    
+    result[i].more = 0;
+
+    result[i].ypresp_all_u.val.stat = YP_NOMORE;
+    result[i].ypresp_all_u.val.key.keydat_len = 0;
+    result[i].ypresp_all_u.val.key.keydat_val = NULL;
+    result[i].ypresp_all_u.val.val.valdat_len = 0;
+    result[i].ypresp_all_u.val.val.valdat_val = NULL;
+
+#endif
+
+    return result;
 }
 
 ypresp_master *ypproc_master_2_svc(ypreq_nokey *nokey, struct svc_req *rqstp)
@@ -320,14 +466,39 @@ ypresp_master *ypproc_master_2_svc(ypreq_nokey *nokey, struct svc_req *rqstp)
 	fprintf(stderr, "\tmap = \"%s\"\n", nokey->map);
     }
 
+    result.peer = "";
 
     if (!is_valid_domain(nokey->domain))
 	result.stat = YP_NODOM;
     else
-	result.stat = YP_BADDB;
+    {
+	GDBM_FILE dbp = open_database(nokey->domain, nokey->map);
+	if (dbp == NULL)
+	    result.stat = YP_NOMAP;
+	else
+	{
+	    datum key, val;
+
+	    key.dsize = sizeof("YP_MASTER_NAME")-1;
+	    key.dptr = "YP_MASTER_NAME";
+	    
+	    val = gdbm_fetch(dbp, key);
+	    if (val.dptr == NULL)
+	    {
+		/* No YP_MASTER_NAME record in map? Assume we are Master */
+		static char hostbuf[1025];
+
+		gethostname(hostbuf, sizeof(hostbuf)-1);
+		result.peer = hostbuf;
+	    }
+	    else
+		result.peer = val.dptr;
+
+	    result.stat = YP_TRUE;
+	    gdbm_close(dbp);
+	}
+    }
 
-    result.peer = "foobar.fubar.com";
-    
     return &result;
 }
 
@@ -342,20 +513,78 @@ ypresp_order *ypproc_order_2_svc(ypreq_nokey *nokey, struct svc_req *rqstp)
 	fprintf(stderr, "\tmap = \"%s\"\n", nokey->map);
     }
 
+    result.ordernum = 0;
 
     if (!is_valid_domain(nokey->domain))
 	result.stat = YP_NODOM;
     else
-	result.stat = YP_BADDB;
+    {
+	GDBM_FILE dbp = open_database(nokey->domain, nokey->map);
+	if (dbp == NULL)
+	    result.stat = YP_NOMAP;
+	else
+	{
+	    datum key, val;
+
+	    key.dsize = sizeof("YP_LAST_MODIFIED")-1;
+	    key.dptr = "YP_LAST_MODIFIED";
+	    
+	    val = gdbm_fetch(dbp, key);
+	    if (val.dptr == NULL)
+	    {
+		/* No YP_LAST_MODIFIED record in map? Use DTM timestamp.. */
+		result.ordernum = get_dtm(nokey->domain, nokey->map);
+	    }
+	    else
+		result.ordernum = atoi(val.dptr);
+
+	    result.stat = YP_TRUE;
+	    gdbm_close(dbp);
+	}
+    }
 
-    result.ordernum = 42;
-    
     return &result;
 }
 
+
+static void free_maplist(ypmaplist *mlp)
+{
+    ypmaplist *next;
+
+    while (mlp != NULL)
+    {
+	next = mlp->next;
+	free(mlp->map);
+	free(mlp);
+	mlp = next;
+    }
+}
+
+static int add_maplist(ypmaplist **mlhp, char *map)
+{
+    ypmaplist *mlp;
+
+    mlp = malloc(sizeof(*mlp));
+    if (mlp == NULL)
+	return -1;
+
+    mlp->map = strdup(map);
+    if (mlp->map == NULL)
+    {
+	free(mlp);
+	return -1;
+    }
+
+    mlp->next = *mlhp;
+    *mlhp = mlp;
+
+    return 0;
+}
+
 ypresp_maplist *ypproc_maplist_2_svc(domainname *name, struct svc_req *rqstp)
 {
     static ypresp_maplist result;
+    
 
     if (debug_flag)
     {
@@ -363,13 +592,30 @@ ypresp_maplist *ypproc_maplist_2_svc(domainname *name, struct svc_req *rqstp)
 	fprintf(stderr, "\tdomain = \"%s\"\n", *name);
     }
 
+    if (result.maps)
+	free_maplist(result.maps);
+    
+    result.maps = NULL;
 
     if (!is_valid_domain(*name))
 	result.stat = YP_NODOM;
     else
-	result.stat = YP_BADDB;
+    {
+	DIR *dp;
 
-    result.maps = NULL;
+	dp = opendir(path_ypdb);
+	if (dp == NULL)
+	    result.stat = YP_BADDB;
+	else
+	{
+	    struct dirent *dep;
+	    
+	    while (dep = readdir(dp))
+		add_maplist(&result.maps, dep->d_name);
+	}
+
+	closedir(dp);
+    }
     
     return &result;
 }
